// Copyright 2012-2014 Masanori Morise. All Rights Reserved.
// Author: mmorise [at] yamanashi.ac.jp (Masanori Morise)
//
// Test program for WORLD 0.1.2 (2012/08/19)
// Test program for WORLD 0.1.3 (2013/07/26)
// Test program for WORLD 0.1.4 (2014/04/29)

// test.exe input.wav outout.wav f0 spec
// input.wav  : Input file
// output.wav : Output file
// f0         : F0 scaling (a positive number)
// spec       : Formant shift (a positive number)
// flag       : 0: Excitation signal, 1: Aperiodicity (same as TANDEM-STRAIGHT)

// Memo: This is my dream.
// D4C (Definitive Decomposition Done Dirt Cheap) will be implemented
// in the next version of WORLD. D4C outputs the aperiodic parameter
// similar to the AperiodicityRatio().

// Announcement:
// Star() was removed from the project.

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#if (defined (__WIN32__) || defined (_WIN32)) && !defined (__MINGW32__)
#include <conio.h>
#include <windows.h>
//only for non mingw
#pragma warning(disable : 4996)
#pragma comment(lib, "winmm.lib")
#endif
#if (defined (__linux__) || defined(__CYGWIN__))
// Linux porting section: include necessary header files under Linux system
#include <stdint.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#ifndef ECHOPRT
#define ECHOPRT ECHOE
#endif
#endif

#include "./aperiodicity.h"
#include "./dio.h"
#include "./matlabfunctions.h"
#include "./platinum.h"
// #include "./star.h"
#include "./cheaptrick.h"  // This is the new function.
#include "./stonemask.h"
#include "./synthesis.h"
#include "./synthesisfromaperiodicity.h"



// Frame shift [msec]
#define FRAMEPERIOD 5.0

#if (defined (__linux__) || defined(__CYGWIN__))
// Linux porting section: implement timeGetTime() by gettimeofday(),
//                        implement getch() by termios
#ifndef DWORD
#define DWORD uint32_t
#endif
DWORD timeGetTime() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  DWORD ret = static_cast<DWORD>(tv.tv_usec / 1000 + tv.tv_sec * 1000);
  return ret;
}
int getch() {
  int ch;
  struct termios oldt, newt;

  tcgetattr(STDIN_FILENO, &oldt);
  memcpy(&newt, &oldt, sizeof(newt));
  newt.c_lflag &=
    ~(ECHO | ICANON | ECHOE | ECHOK | ECHONL | ECHOPRT | ECHOKE | ICRNL);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  ch = getchar();
  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);

  return ch;
}
#endif

namespace {
bool CheckLoadedFile(double *x, int fs, int nbit, int x_length) {
  if (x == NULL) {
    printf("error: File not found.\n");
    getch();
    return false;
  }

  printf("File information\n");
  printf("Sampling : %d Hz %d Bit\n", fs, nbit);
  printf("Length %d [sample]\n", x_length);
  printf("Length %f [sec]\n", static_cast<double>(x_length) / fs);
  return true;
}

void F0Estimation(double *x, int x_length, int fs, int f0_length, double *f0,
    double *time_axis) {
  double *refined_f0 = new double[f0_length];

  DioOption option;  // WORLD 0.1.1 and later
  InitializeDioOption(&option);  // Initialize the option
  // modification of the option
  option.frame_period = FRAMEPERIOD;
  // Following value represents the ratio for downsampling.
  // The signal is downsampled to fs / speed Hz.
  option.speed = 12;
  option.f0_floor = 80.0;
  option.f0_ceil = 640.0;
  option.channels_in_octave = 2.0;

  printf("\nAnalysis\n");
  DWORD elapsed_time = timeGetTime();
  Dio(x, x_length, fs, option, time_axis, f0);
  printf("DIO: %d [msec]\n", timeGetTime() - elapsed_time);

  // StoneMask is carried out to improve the estimation performance.
  elapsed_time = timeGetTime();
  StoneMask(x, x_length, fs, time_axis, f0, f0_length, refined_f0);
  printf("StoneMask: %d [msec]\n", timeGetTime() - elapsed_time);

  for (int i = 0; i < f0_length; ++i) f0[i] = refined_f0[i];

  delete[] refined_f0;
  return;
}

void SpectralEnvelopeEstimation(double *x, int x_length, int fs,
    double *time_axis, double *f0, int f0_length, double **spectrogram) {
  DWORD elapsed_time = timeGetTime();
  CheapTrick(x, x_length, fs, time_axis, f0, f0_length, spectrogram);
  printf("CheapTrick: %d [msec]\n", timeGetTime() - elapsed_time);
}

void AperiodicParameterEstimation(double *x, int x_length, int fs,
    double *time_axis, double *f0, int f0_length, double **spectrogram,
    bool tandem_flag, double **residual_spectrogram, double **aperiodicity) {
  int fft_size = GetFFTSizeForCheapTrick(fs);
  DWORD elapsed_time = timeGetTime();
  if (tandem_flag == false) {
    // Excitation signal.
    elapsed_time = timeGetTime();
    Platinum(x, x_length, fs, time_axis, f0, f0_length,
        spectrogram, fft_size, residual_spectrogram);
    printf("PLATINUM: %d [msec]\n", timeGetTime() - elapsed_time);
  } else {
    // Aperiodicity.
    elapsed_time = timeGetTime();
    AperiodicityRatio(x, x_length, fs, f0, f0_length, time_axis, fft_size,
      aperiodicity);
    printf("TANDEM_AP: %d [msec]\n", timeGetTime() - elapsed_time);
  }
}

void ParameterModification(int argc, char *argv[], int fs, double *f0,
    int f0_length, double **spectrogram) {
  int fft_size = GetFFTSizeForCheapTrick(fs);
  // F0 scaling
  if (argc >= 4) {
    double shift = atof(argv[3]);
    for (int i = 0; i < f0_length; ++i) f0[i] *= shift;
  }
  // Spectral stretching
  if (argc >= 5) {
    double ratio = atof(argv[4]);
    double *freq_axis1 = new double[fft_size];
    double *freq_axis2 = new double[fft_size];
    double *spectrum1 = new double[fft_size];
    double *spectrum2 = new double[fft_size];

    for (int i = 0; i <= fft_size / 2; ++i) {
      freq_axis1[i] = ratio * i / fft_size * fs;
      freq_axis2[i] = static_cast<double>(i) / fft_size * fs;
    }
    for (int i = 0; i < f0_length; ++i) {
      for (int j = 0; j <= fft_size / 2; ++j)
        spectrum1[j] = log(spectrogram[i][j]);
      interp1(freq_axis1, spectrum1, fft_size / 2 + 1, freq_axis2,
        fft_size / 2 + 1, spectrum2);
      for (int j = 0; j <= fft_size / 2; ++j)
        spectrogram[i][j] = exp(spectrum2[j]);
      if (ratio < 1.0) {
        for (int j = static_cast<int>(fft_size / 2.0 * ratio);
            j <= fft_size / 2; ++j)
          spectrogram[i][j] =
          spectrogram[i][static_cast<int>(fft_size / 2.0 * ratio) - 1];
      }
    }
    delete[] spectrum1;
    delete[] spectrum2;
    delete[] freq_axis1;
    delete[] freq_axis2;
  }
}

void WaveformSynthesis(double *f0, int f0_length, double **spectrogram,
    double **residual_spectrogram, double **aperiodicity, int fft_size,
    double frame_period, int fs, bool tandem_flag, int y_length,
    double *y) {
  DWORD elapsedTime;
  if (tandem_flag == false) {
    // Synthesis
    printf("\nSynthesis\n");
    elapsedTime = timeGetTime();
    Synthesis(f0, f0_length, spectrogram, residual_spectrogram,
        fft_size, FRAMEPERIOD, fs, y_length, y);
    printf("WORLD: %d [msec]\n", timeGetTime() - elapsedTime);
  } else {
    // Synthesis by the aperiodicity
    printf("\nSynthesis\n");
    elapsedTime = timeGetTime();
    SynthesisFromAperiodicity(f0, f0_length, spectrogram, aperiodicity,
        fft_size, FRAMEPERIOD, fs, y_length, y);
    printf("WORLD: %d [msec]\n", timeGetTime() - elapsedTime);
  }
}

}  // namespace


//-----------------------------------------------------------------------------
// Test program.
// test.exe input.wav outout.wav f0 spec flag
// input.wav  : argv[1] Input file
// output.wav : argv[2] Output file
// f0         : argv[3] F0 scaling (a positive number)
// spec       : argv[4] Formant shift (a positive number)
// flag       : argv[5] Excitation signal or Aperiodicity(0 or 1)
//-----------------------------------------------------------------------------
int main(int argc, char *argv[]) {
  if (argc != 2 && argc != 3 && argc != 4 && argc != 5 && argc != 6) {
    printf("usage: world_test infile.wav outfile.wav [pitchshift] [timbrescale] [tandem_flag]\n");

    return 0;
  }
  int fs, nbit, x_length;
  double *x = wavread(argv[1], &fs, &nbit, &x_length);

  if (CheckLoadedFile(x, fs, nbit, x_length) == false) {
    printf("error: File not found.\n");
    return 0;
  }

  // Allocate memories
  // The number of samples for F0
  int f0_length = GetSamplesForDIO(fs, x_length, FRAMEPERIOD);
  double *f0 = new double[f0_length];
  double *time_axis = new double[f0_length];

  // FFT size for CheapTrick, PLATINUM and Aperiodicity
  int fft_size = GetFFTSizeForCheapTrick(fs);
  double **spectrogram = new double *[f0_length];
  double **residual_spectrogram = new double *[f0_length];
  double **aperiodicity = new double *[f0_length];  // for aperiodicity
  for (int i = 0; i < f0_length; ++i) {
    spectrogram[i] = new double[fft_size / 2 + 1];
    residual_spectrogram[i] = new double[fft_size + 1];  // miss??
    aperiodicity[i] = new double[fft_size / 2 + 1];
  }

  // F0 estimation
  //-------------------------------
  F0Estimation(x, x_length, fs, f0_length, f0, time_axis);

  // Spectral envelope estimation
  SpectralEnvelopeEstimation(x, x_length, fs, time_axis, f0, f0_length,
      spectrogram);

  bool tandem_flag = false;
  if (argc == 6) if (argv[5][0] == '1') tandem_flag = true;
  // Aperiodic parameter (residual_spectrogram by PLATINUM() or
  // aperiodicity by AperiodicityRatio())
  AperiodicParameterEstimation(x, x_length, fs, time_axis, f0, f0_length,
      spectrogram, tandem_flag, residual_spectrogram, aperiodicity);

  // Note that F0 must not be changed until all parameters are estimated.
  ParameterModification(argc, argv, fs, f0, f0_length, spectrogram);

  // The length of the output waveform
  int y_length =
    static_cast<int>((f0_length - 1) * FRAMEPERIOD / 1000.0 * fs) + 1;
  double *y = new double[y_length];

  // Synthesis
  WaveformSynthesis(f0, f0_length, spectrogram, residual_spectrogram,
      aperiodicity, fft_size, FRAMEPERIOD, fs, tandem_flag, y_length, y);

  // Output
  wavwrite(y, y_length, fs, 16, argv[2]);

  printf("complete.\n");

  delete[] x;
  delete[] time_axis;
  delete[] f0;
  delete[] y;
  for (int i = 0; i < f0_length; ++i) {
    delete[] spectrogram[i];
    delete[] residual_spectrogram[i];
    delete[] aperiodicity[i];
  }
  delete[] spectrogram;
  delete[] residual_spectrogram;
  delete[] aperiodicity;

  return 0;
}
